/*-
 * Copyright (c) 1999-2000 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: ipa.c,v 1.1.2.5 2000/12/01 10:09:47 simon Exp $";
#endif /* !lint */

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <grp.h>
#include <pwd.h>

#include "ipa.h"

#include "cmd.h"
#include "common.h"
#include "config.h"
#include "kipfil.h"
#include "kipfw.h"
#include "path.h"
#include "rules.h"



uid_t		myuid;
gid_t		mygid;
int		debug = 0; /* 1 if switch -d was used */

static char	*envprogname;
static int	pidfd;
static char	*pidfilename = PIDFILE;
static const char ipa_ident[] = IPA_IDENT;
static const char ipa_msg[] = IPA_MSG;

static void	show_usage(void), show_version(void);
static void	daemon_kill(int);
static int	daemon_init(void);
static int	lock_pidfile(void);
static int	set_sighandlers(int);
static void	abnormalterm(void);

#ifdef __GNUC__
static void	fprintf_error(const char *, ...) __attribute__ ((format (printf, 1, 2)));
#else
static void	fprintf_error(const char *, ...);
#endif


int
main(int argc, char *argv[])
{
	int	opt;
	int	testconfig = 0; /* 1 if switch -t was used */

	envprogname = argv[0];	/* save program name */
	opterr = 0;		/* don't allow getopt() print own messages */
	umask(0007);

	while ( (opt = getopt(argc, argv, "dtvhf:k:")) != -1)
		switch (opt) {
		case 'v':
			show_version();
			return 0;
		case 'h':
			show_usage();
			return 0;
		case 'd':
			debug = 1;
			break;
		case 't':
			testconfig = 1;
			break;
		case 'f':
			cfgfilename = optarg;
			break;
		case 'k':
			if (strcmp(optarg, "shutdown") == 0)
				daemon_kill(SIGTERM);
			if (strcmp(optarg, "kill") == 0)
				daemon_kill(SIGKILL);
			if (strcmp(optarg, "reconfigure") == 0)
				daemon_kill(SIGHUP);
			fprintf(stderr, "%s: invalid parameter \"%s\" for switch -k\n", argv[0], optarg);
			return 1;
		case '?':
			fprintf(stderr, "%s: invalid switch -%c\n", argv[0], optopt);
			return 1;
		default:
			fprintf(stderr, "%s: getopt: %s\n", envprogname, strerror(errno));
			return 1;
		}

	if (optind < argc) {
		for (; optind < argc; ++optind)
			fprintf(stderr, "%s: non-switch argument \"%s\"\n", envprogname, argv[optind]);
		return 1;
	}

	if (testconfig) {
		if (parse_config(TEST_PARSING) < 0)
			return 1;
		show_config();
		return 0;
	}

	myuid = getuid();
	mygid = getgid();

	if (myuid != 0) {
		fprintf(stderr, "%s: %s should be run by root\n", argv[0], ipa_msg);
		return 1;
	}
	if (lock_pidfile() < 0)
		return 1;
	openlog(IPA_IDENT, LOG_PID|(debug ? LOG_PERROR : LOG_CONS), LOG_USER);
	if (!debug) {
		if (daemon_init() < 0) {
			syslog(LOG_ERR, "can't force daemon regime");
			abnormalterm();
		}
		if (lock_pidfile() < 0)
			abnormalterm();
	}

	syslog(LOG_INFO, "------------------------------------------------");
	syslog(LOG_INFO, "%s/%s daemon started by UID %d GID %d",
	    ipa_ident, Version, myuid, mygid);
	syslog(LOG_INFO, "use configuration file %s, parsing...", cfgfilename);
	if (parse_config(STARTUP_PARSING) < 0)
		abnormalterm();

	loaded_rules_info();

	if (build_db_regexes() < 0)
		abnormalterm();
	if (init_db() < 0)
		abnormalterm();

	if (use_ipfw)
		if (kipfw_init() < 0)
			abnormalterm();
	if (use_ipfil)
		if (kipfil_init() < 0)
			abnormalterm();

	if (set_sighandlers(0) < 0)
		abnormalterm();

	/* Startup */
	if (startup_global.cmd) {
		syslog(LOG_INFO, "run startup(global) commands...");
		if (exec_cmd_list(startup_global.cmd, startup_global.ncmd, "startup(global)") < 0)
			abnormalterm();
		free_cmd_list(startup_global.cmd, startup_global.ncmd);
	}

	if (run_rules_rc(0) < 0)
		abnormalterm();

	if (set_sighandlers(1) < 0) /* set SIGCHLD handler */
		abnormalterm();

	/* Working... */
	if (run_ipac() < 0) {
		syslog(LOG_ERR, "can't make IP accounting");
		abnormalterm();
	}

	if (set_sighandlers(2) < 0) /* set SIGCHLD handler */
		abnormalterm();

	/* Shutdown */
	if (run_rules_rc(1) < 0)
		abnormalterm();
	if (shutdown_global.cmd) {
		syslog(LOG_INFO, "run shutdown(global) commands...");
		if (exec_cmd_list(shutdown_global.cmd, shutdown_global.ncmd, "shutdown(global)") < 0)
			abnormalterm();
	}

	syslog(LOG_INFO, "%s/%s daemon stoped", ipa_ident, Version);
	return 0;
}

/*
 * Output message to stderr stream, if errno != 0 then
 * output error message.
 */
void
fprintf_error(const char *message, ...)
{
	int	errno_save = errno;
	va_list	ap;

	va_start(ap, message);
	fflush(stdout);
	fprintf(stderr, "%s: ", envprogname);
	vfprintf(stderr, message, ap);
	if (errno_save)
		fprintf(stderr, ": %s\n", strerror(errno_save));
	else
		fprintf(stderr, "\n");
	va_end(ap);
}

/*
 * Init daemon: ignore signals TTOU, TTIN and TSTP, fork() new process,
 * chdir() to /, close() all available descriptors.
 */
int
daemon_init(void)
{
	pid_t		childpid;
	int		fd;
	struct rlimit	flim;

	if (signal(SIGTTOU, SIG_IGN) < 0) {
		syslog(LOG_ERR, "signal(SIGTTOU, SIG_IGN): %m");
		return -1;
	}
	if (signal(SIGTTIN, SIG_IGN) < 0) {
		syslog(LOG_ERR, "signal(SIGTTIN, SIG_IGN): %m");
		return -1;
	}
	if (signal(SIGTSTP, SIG_IGN) < 0) {
		syslog(LOG_ERR, "signal(SIGTSTP, SIG_IGN): %m");
		return -1;
	}

	if ( (childpid = fork()) < 0) {
		syslog(LOG_ERR, "fork: %m");
		return -1;
	}
	if (childpid != 0)
		_exit(0);	/* parent goes */
	/* child continues */
	setsid();
	chdir("/");
	getrlimit(RLIMIT_NOFILE, &flim);
	for (fd = 0; fd < flim.rlim_max; ++fd)
		if (fd != pidfd)
			close(fd);
	errno = 0;
	return 0;
}

/*
 * Lock file pidfilename against simultaneouse run copies and
 * store PID of process to this file.
 *
 * This function should be called twise: first call locks file pidfilename
 * and stores PID of process, second call just stores PID of process (it's
 * necessary, because daemon_init() fork() new process).
 */
int
lock_pidfile(void)
{
	int		val;
	static FILE	*fp = NULL;

	if (!fp) {
		/* first call, don't use syslog */
		if ( ((pidfd = open(pidfilename, O_RDWR|O_CREAT, 0644)) < 0) ||
		    ((fp = fdopen(pidfd, "r+")) == NULL) ) {
			fprintf_error("can't open or create file %s", pidfilename);
			return -1;
		}
		if (flock(pidfd, LOCK_EX|LOCK_NB) < 0) {
			int	errno_save = errno, other_pid;

			if (errno != EWOULDBLOCK) {
				fprintf_error("flock(%s)", pidfilename);
				return -1;
			}
			if (fscanf(fp, "%d", &other_pid) != 1)
				fprintf(stderr, "%s: can't lock file %s: %s\n",
				    envprogname, pidfilename, strerror(errno_save));
			else
				fprintf(stderr, "%s: can't lock file %s, it may be locked by PID %d: %s\n",
				    envprogname, pidfilename, other_pid, strerror(errno_save));
			return -1;
		}
		if ( (val = fcntl(pidfd, F_GETFD, 0)) < 0) {
			fprintf_error("fcntl(%s, F_GETFD)", pidfilename);
			return -1;
		}
		val |= FD_CLOEXEC;
		if (fcntl(pidfd, F_SETFD, val) < 0) {
			fprintf_error("fcntl(%s, F_SETFD)", pidfilename);
			return -1;
		}
		if (fseek(fp, 0L, SEEK_SET) < 0) {
			fprintf_error("fseek(%s, 0, SEEK_SET)", pidfilename);
			return -1;
		}
		if (fprintf(fp, "%d\n", getpid()) < 0) {
			fprintf_error("fprintf(%s)", pidfilename);
			return -1;
		}
		if (fflush(fp) != 0) {
			fprintf_error("fflush(%s)", pidfilename);
			return -1;
		}
		if (ftruncate(pidfd, ftello(fp)) < 0) {
			fprintf_error("ftruncate(%s)", pidfilename);
			return -1;
		}
	} else {
		/* next call, use syslog */
		if (fseek(fp, 0L, SEEK_SET) < 0) {
			syslog(LOG_ERR, "fseek(%s, 0, SEEK_SET): %m", pidfilename);
			return -1;
		}
		if (fprintf(fp, "%d\n", getpid()) < 0) {
			syslog(LOG_ERR, "fprintf(%s): %m", pidfilename);
			return -1;
		}
		if (fflush(fp) != 0) {
			syslog(LOG_ERR, "fflush(%s): %m", pidfilename);
			return -1;
		}
		if (ftruncate(pidfd, ftello(fp)) < 0) {
			syslog(LOG_ERR, "ftruncate(%s): %m", pidfilename);
			return -1;
		}
	}
	return 0;
}

/*
 * Set handlers for signals HUP, ALRM, CHLD, TERM and INT (if -d switch
 * was specified in command line and daemon is run as foreground
 * process).
 */
int
set_sighandlers(int phase)
{
	struct sigaction	act;

	act.sa_flags = 0;

	sigemptyset(&act.sa_mask);
	if (debug)
		sigaddset(&act.sa_mask, SIGINT);
	sigaddset(&act.sa_mask, SIGHUP);
	sigaddset(&act.sa_mask, SIGTERM);
	sigaddset(&act.sa_mask, SIGALRM);
	sigaddset(&act.sa_mask, SIGCHLD);

	switch (phase) {
	case 0:
		act.sa_handler = end_work;
		if (debug) {
			if (sigaction(SIGINT, &act, (struct sigaction *)0) < 0) {
				syslog(LOG_ERR, "sigaction(SIGINT): %m");
				return -1;
			}
		}

		if (sigaction(SIGTERM, &act, (struct sigaction *)0) < 0) {
			syslog(LOG_ERR, "sigaction(SIGTERM): %m");
			return -1;
		}

		act.sa_handler = reconfigure;
		if (sigaction(SIGHUP, &act, (struct sigaction *)0) < 0) {
			syslog(LOG_ERR, "sigaction(SIGHUP): %m");
			return -1;
		}

		act.sa_handler = sig_alrm;
		if (sigaction(SIGALRM, &act, (struct sigaction *)0) < 0) {
			syslog(LOG_ERR, "sigaction(SIGALRM): %m");
			return -1;
		}
		break;
	case 1:
		act.sa_handler = sig_chld;
		if (sigaction(SIGCHLD, &act, (struct sigaction *)0) < 0) {
			syslog(LOG_ERR, "sigaction(SIGCHLD): %m");
			return -1;
		}
		break;
	case 2:
		if (signal(SIGCHLD, SIG_DFL) < 0) {
			syslog(LOG_ERR, "signal(SIGCHLD, SIG_DFL): %m");
			return -1;
		}
		break;
	default:
		syslog(LOG_ERR, "set_sighandlers: unknown phase number %d", phase);
		return -1;
	}
	return 0;
}

/*
 * Send signal signo to running copy of daemon. PID of daemon is taken
 * from file pidfilename.
 *
 * XXX daemon_kill() doesn't check if daemon is realy was run.
 */
void
daemon_kill(int signo)
{
	FILE	*fp;
	int	pid;

	if ( (fp = fopen(pidfilename, "r")) == NULL) {
		fprintf_error("fopen(%s)", pidfilename);
		exit(1);
	}
	if (fscanf(fp, "%d", &pid) != 1) {
		fprintf_error("fscanf(%s, %%d) failed", pidfilename);
		exit(1);
	}
	printf("Sending signal %d '%s' to PID %d...", signo, sys_signame[signo], pid);

	if (kill(pid, signo) < 0) {
		fprintf_error("kill(%d, '%s')", pid, sys_signame[signo]);
		exit(1);
	}
	printf(" done\n");
	exit(0);
}

/*
 * Serious errors aoccured, log message and exit() from daemon.
 */
void
abnormalterm(void)
{
	syslog(LOG_ERR, "abnormal termination");
	exit(1);
}

/*
 * Output version number (-v and -h switch).
 */
void
show_version(void)
{
	printf("%s/%s\n", ipa_ident, Version);
}

/*
 * Output help message (-h switch).
 */
void
show_usage(void)
{
	char	*ptr;
	
	if ( (ptr = strrchr(envprogname, '/')) != NULL)
		++ptr;
	else
		ptr = envprogname;

	show_version();
	printf("%s\n\
Usage: %s [-tdvh] [-f <config-file>] [-k <signal>]\n\
   -t\tParse configuration file, output its content and exit.\n\
   -d\tDon't switch to daemon regime and write all messages to stderr\n\
   \tas well to the syslog.\n\
   -f <config-file>\t\n\
   \tUse given <config-file> instead of using default configuration file\n\
   \t%s\n\
   -k shutdown|reconfigure|kill\t\n\
   \tSend signal to running copy:\n\
   \t* shutdown - send signal TERM (cause shutdowning)\n\
   \t* reconfigure - send signal HUP (cause rereading configuration file)\n\
   \t* kill - send signal KILL\n\
   -v\tShow version number and exit.\n\
   -h\tOutput this help message.\n", ipa_msg, ptr, CFGFILE);
}
