/*-
 * Copyright (c) 1999-2000 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: rules.c,v 1.1.2.5 2000/12/01 10:09:47 simon Exp $";
#endif /* !lint */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <netinet/ip_fw.h>
#include <dirent.h>
#include <unistd.h>
#include <varargs.h>
#include <regex.h>

#include "rules.h"

#include "cmd.h"
#include "common.h"
#include "config.h"
#include "debug.h"
#include "ipa.h"
#include "kipfil.h"
#include "kipfw.h"
#include "path.h"



/* PERMMASK allows to get just permissions bit from st_mode field
   PERMMAS(x) = x & 0777 */
#define PERMMASK(x)	((x) & (S_IRWXU|S_IRWXG|S_IRWXO))

char		*acctdir = ACCTDIR;	/* main database directory */

u_long		db_uid, db_gid;
u_int		db_fileperm, db_dirperm;

struct ipac	*ipac = NULL;		/* ptr on accounting rules array */
u_int		nipac = 0;		/* #rules */

/* use_ipfw (use_ipfil) are equal to 1 if "ipfw" ("ipfil") parameter
   was used in some "rule" section */
int		use_ipfw = 0, use_ipfil = 0;

/* *_global variables are equal to parameters in "global" section */
char		*acl_global;
short		use_worktime_global = 0;
struct worktime	worktime_global[7];
struct commands	startup_global;
struct commands	shutdown_global;

static time_t	curr_time;	/* current UTC time in seconds */
static struct tm curr_tm;	/* current localtime */
static time_t	newday_time;	/* UTC time in seconds of next day */

static u_int	sleep_seconds;	/* how many seconds to sleep */
static int	newday_is_comming = 0;	/* ==1 if new day came */
static int	firstcall = 1;	/* ==0 if we pass one loop in run_ipac() */
static int	process_mon = 0, /* current month */
		process_year = 0;/* current year */
static regex_t	reg_db_entry, reg_limit_line1, reg_limit_timestamp;

static volatile sig_atomic_t	goingdown_flag = 0, alarm_flag = 0, wakeup_flag = 0,
				reconfigure_flag = 0, time_changed_flag = 0,
				sig_chld_flag = 0;


static int	do_reconfigure(void);
static int	free_ipac_list(struct ipac *, u_int);

static int	update_limits(struct ipac *);
static void	add_chunk_to_rule_limits(struct ipac *, u_quad_t);
static void	sub_chunk_from_rule_limits(struct ipac *, u_quad_t);
static time_t	check_rule_limits(struct ipac *, time_t *);
static int	init_limits(int);

static int	check_db_file_ownership(const char *, struct stat *);
static int	check_db_dir_ownership(const char *, struct stat *);
static int	chown_db(const char *);
static int	mkdir_db(const char *);
static FILE	* create_db_file(const char *);
static int	Ftruncate_Seek(FILE *, long, const char *);

static int	append_limit_z_ts(struct limit *);
static int	append_limit_e_ts(struct limit *);
static int	restart_limit(struct ipac *, struct limit *);

static time_t	mk_time_exp_upto(time_t, char);


/*
 * Update current record for *acp accounting rule in data base.
 * value is equal to new value of counter.
 */
static int
update_db(struct ipac *acp, u_quad_t value)
{
	if (fseek(acp->fp, -DBCOUNTEROFF, SEEK_END) < 0) {
		syslog(LOG_ERR, "fseek(%s, -%d, SEEK_END): %m",
		    acp->filename, DBCOUNTEROFF);
		return -1;
	}
	/* update second timestamp and counter */
	if (fprintf(acp->fp, "%02d:%02d:%02d %020qu",
	    curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec, value) != DBCOUNTERSIZE) {
		syslog(LOG_ERR, "fprintf(%s): %m", acp->filename);
		return -1;
	}
	if (fflush(acp->fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", acp->filename);
		return -1;
	}
	return 0;
}

/*
 * Append new record for *acp accounting rule in data base.
 * value is equal to value of new record.
 */
static int
append_db(struct ipac *acp, u_quad_t value)
{
	if (fseek(acp->fp, 0L, SEEK_END) < 0) {
		syslog(LOG_ERR, "fseek(%s, 0, SEEK_END): %m", acp->filename);
		return -1;
	}
	if (fprintf(acp->fp, "%02d/%02d:%02d:%02d-%02d:%02d:%02d %020qu\n",
	    curr_tm.tm_mday,
	    curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec,
	    curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec,
	    value) != DBRECORDSIZE) {
		syslog(LOG_ERR, "fprintf(%s): %m", acp->filename);
		return -1;
	}
	/* newrec_time will be checked if "append_db_time" parameter was
	   specified in the configuration file for *acp rule */
	acp->newrec_time = curr_time + acp->append_db_time;
	if (fflush(acp->fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", acp->filename);
		return -1;
	}
	return 0;
}

/*
 * Add chunk to counter of *acp rule and to counters of its limits.
 * Check if there is some "subtracted" bytes. Check for counter overflowing.
 */
static int
add_chunk_to_rule(struct ipac *acp, u_quad_t chunk)
{
	if (acp->bcnt_sub >= chunk)
		acp->bcnt_sub -= chunk;
	else {
		chunk -= acp->bcnt_sub;
		acp->bcnt_sub = 0;
		if (acp->bcnt < UQUAD_MAX - chunk)
			acp->bcnt += chunk;
		else {
			if (update_db(acp, UQUAD_MAX) < 0)
				return -1;
			acp->bcnt = UQUAD_MAX - acp->bcnt;
			acp->bcnt = chunk - acp->bcnt;
			if (append_db(acp, acp->bcnt) < 0)
				return -1;
		}
	}
	if (acp->limit)
		add_chunk_to_rule_limits(acp, chunk);
	return 0;
}

/*
 * Do IP accounting for rule *acp from IP Firewall table.
 */
static int
do_ipfwac_for_rule(struct ipac *acp)
{
	u_int		i, j;
	int		seen_flag;
	int		prevnum = -1;	/* real value of number is u_short type */
	u_int		subnumber = 0;	/* initial value isn't used */
	u_quad_t	chunk = 0;	/* initial value isn't used */
	u_quad_t	maxchunk = acp->maxchunk;
	struct ipfwac	*fwacp = acp->ipfwac;
	struct ip_fw	*kfwp = kipfw;

	for (i = j = 0; i < acp->nipfwac; ++fwacp, ++i) {
		for (seen_flag = 0; !seen_flag && j < nkipfw; ++kfwp, ++j) {
			if (kfwp->fw_number > fwacp->number)
				break;
			if (prevnum == kfwp->fw_number)
				++subnumber;
			else {
				prevnum = kfwp->fw_number;
				subnumber = 0;
			}
			if (kfwp->fw_number == fwacp->number &&
			    subnumber == fwacp->subnumber) {
				/* found IPFW rule in kernel table */
				seen_flag = 1;
				if (fwacp->set == 0) {
					fwacp->bcnt_old = kfwp->fw_bcnt;
					fwacp->set = 1;
					if (!firstcall)
						syslog(LOG_WARNING, "ipfw rule %hu.%u was added to IPFW kernel table (rule %s)",
						    fwacp->number, fwacp->subnumber, acp->rulename);
				} else if (fwacp->bcnt_old != kfwp->fw_bcnt) {
					if (fwacp->bcnt_old < kfwp->fw_bcnt)
						chunk = kfwp->fw_bcnt - fwacp->bcnt_old;
					else
						chunk += UQUAD_MAX - (fwacp->bcnt_old - kfwp->fw_bcnt);
					fwacp->bcnt_old = kfwp->fw_bcnt;
					if (maxchunk != 0) {
						if (chunk > maxchunk) {
							syslog(LOG_WARNING, "Was ipfw rule %hu.%u flushed? Rule %s maxchunk = %s, current chunk = %s",
							    fwacp->number, fwacp->subnumber, acp->rulename,
							    show_size(maxchunk), show_size2(chunk));
							continue;
						}
					}
					if (fwacp->action == ADD) {
						if (add_chunk_to_rule(acp, chunk) < 0)
							return -1;
					} else /* fwacp->action == SUB */ {
						if (acp->bcnt >= chunk)
							acp->bcnt -= chunk;
						else {
							if (acp->bcnt_sub < UQUAD_MAX - (chunk - acp->bcnt))
								acp->bcnt_sub += chunk - acp->bcnt;
							else {
								syslog(LOG_ERR, "rule %s: wrong configuration for ipfw rule -%hu.%u, counter overflowed",
								    acp->rulename, fwacp->number, fwacp->subnumber);
								acp->bcnt_sub = UQUAD_MAX;
							}
							acp->bcnt = 0;
						}
						if (acp->limit)
							sub_chunk_from_rule_limits(acp, chunk);
					}
				}
			}
		}
		if (!seen_flag) {
			if (fwacp->set != 0) {
				fwacp->set = 0;
				syslog(LOG_WARNING, "ipfw rule %hu.%u was removed from IPFW kernel table (rule %s)",
				    fwacp->number, fwacp->subnumber, acp->rulename);
			} else if (firstcall)
				syslog(LOG_WARNING, "ipfw rule %hu.%u doesn't exist in IPFW kernel table (rule %s)",
				    fwacp->number, fwacp->subnumber, acp->rulename);
		}
	}
	if (update_db(acp, acp->bcnt) < 0)
		return -1;
	return update_limits(acp);
}

/*
 * Do IP accounting for rule *acp from IP Filter accounting table.
 */
static int
do_ipfilac_for_rule(struct ipac *acp)
{
	u_int		i, j;
	u_int		k = 0;		/* initial value isn't used */
	int		seen_group_flag;
	U_QUAD_T	kbcnt;
	U_QUAD_T	chunk = 0;	/* initial value isn't used */
	u_quad_t	maxchunk = acp->maxchunk;
	struct ipfilac		*filacp;
	struct ipfilac_list	*list = NULL; /* initial value isn't used */
	struct kipfil		*kfilp = kipfil;

	for (i = j = 0, filacp = acp->ipfilac; i < acp->nipfilac; ++filacp, ++i) {
		for (seen_group_flag = 0; j < nkipfil; ++kfilp, ++j) {
			if (kfilp->group > filacp->group || seen_group_flag)
				break;
			if (kfilp->group == filacp->group &&
			    kfilp->type == filacp->type) {
				/* found IPFIL group in kernel table */
				seen_group_flag = 1;
				for (k = 0, list = filacp->list; k < filacp->nlist; ++list, ++k) {
					if (list->number < kfilp->size + 1) {
						kbcnt = kfilp->bcnt[list->number - 1];
						if (list->set == 0) {
							list->bcnt_old = kbcnt;
							list->set = 1;
							if (!firstcall)
								syslog(LOG_WARNING, "ipfil accounting rule %c%hu@%u was added to IPFIL kernel table (rule %s)",
								    filacp->type, filacp->group, list->number, acp->rulename);
						} else {
							if (list->bcnt_old != kbcnt) {
								if (list->bcnt_old < kbcnt)
									chunk = kbcnt - list->bcnt_old;
								else
									chunk += U_QUAD_T_MAX - (list->bcnt_old - kbcnt);
								list->bcnt_old = kbcnt;
								if (maxchunk != 0) {
									if (chunk > maxchunk) {
										syslog(LOG_WARNING, "Was ipfil rule %c%hu@%u flushed? Rule %s maxchunk = %s, curremt chunk = %s",
										    filacp->type, filacp->group, list->number, acp->rulename,
										    show_size(maxchunk), show_size2(chunk));
										continue;
									}
								}
								if (list->action == ADD) {
									if (add_chunk_to_rule(acp, chunk) < 0)
										return -1;
								} else /* list->action == SUB */ {
									if (acp->bcnt >= chunk)
										acp->bcnt -= chunk;
									else {
										if (acp->bcnt_sub < UQUAD_MAX - (chunk - acp->bcnt))
											acp->bcnt_sub += chunk - acp->bcnt;
										else {
											syslog(LOG_ERR, "rule %s: wrong configuration for ipfil rule -%c%hu@%u, counter overflowed",
											    acp->rulename, filacp->type, filacp->group, list->number);
											acp->bcnt_sub = UQUAD_MAX;
										}
										acp->bcnt = 0;
									}
									if (acp->limit)
										sub_chunk_from_rule_limits(acp, chunk);
								}
							}
						}
					} else
						/* some of ipfil rules haven't been found */
						break;
				}
			}
		}
		if (seen_group_flag == 0) {
			k = 0;
			list = filacp->list;
		}
		if ((seen_group_flag == 1 && list->number >= kfilp->size + 1) ||
		    seen_group_flag == 0) {
			for (; k < filacp->nlist; ++list, ++k) {
				if (list->set != 0) {
					list->set = 0;
					syslog(LOG_WARNING, "ipfil accounting rule %c%hu@%u was removed from IPFIL kernel table (rule %s)",
					    filacp->type, filacp->group, list->number, acp->rulename);
				} else if (firstcall)
					syslog(LOG_WARNING, "ipfil accounting rule %c%hu@%u doesn't exist in IPFIL kernel table (rule %s)",
					    filacp->type, filacp->group, list->number, acp->rulename);
			}
		}
	}
	if (update_db(acp, acp->bcnt) < 0)
		return -1;
	return update_limits(acp);
}

static int
do_ipac(void)
{
#define UPDATE_SLEEP_SECONDS()				\
	rule_sleep_seconds = acp->wakeup - curr_time;	\
	if (sleep_seconds > rule_sleep_seconds)		\
		sleep_seconds = rule_sleep_seconds

	u_int		i, rule_sleep_seconds;
	time_t		wakeup_limits;
	int		need_check = goingdown_flag | reconfigure_flag |
				     newday_is_comming | time_changed_flag |
				     sig_chld_flag;
	struct ipac	*acp;

	sleep_seconds = DAY;	/* 1 day in seconds */
	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		if (acp->wakeup <= curr_time || need_check) {
			if (acp->ipfwac)
				if (do_ipfwac_for_rule(acp) < 0)
					return -1;
			if (acp->ipfilac)
				if (do_ipfilac_for_rule(acp) < 0)
					return -1;
			if (acp->append_db_time != 0)
				if (acp->newrec_time <= curr_time) {
					if (append_db(acp, 0ULL) < 0)
						return -1;
					acp->bcnt = 0;
				}
			acp->wakeup = curr_time + acp->update_db_time;
			rule_sleep_seconds = acp->update_db_time;
			if (sleep_seconds > acp->update_db_time)
				sleep_seconds = rule_sleep_seconds;
			if (acp->append_db_time != 0)
				if (acp->wakeup > acp->newrec_time) {
					acp->wakeup = acp->newrec_time;
					UPDATE_SLEEP_SECONDS();
				}
			if (acp->limit) {
				if (check_rule_limits(acp, &wakeup_limits) < 0)
					return -1;
				if (wakeup_limits > 0 && acp->wakeup > wakeup_limits) {
					acp->wakeup = wakeup_limits;
					UPDATE_SLEEP_SECONDS();
				}
			}
			if (debug_time > 1)
				syslog(LOG_INFO, "do_ipac: rule %s, new sleep = %u sec.", acp->rulename, rule_sleep_seconds);
		} else {
			UPDATE_SLEEP_SECONDS();
			if (debug_time > 1)
				syslog(LOG_INFO, "do_ipac: rule %s, rest sleep = %u sec.", acp->rulename, rule_sleep_seconds);
		}
	}
	sig_chld_flag = 0;
	return 0;
}

static int
newday_came(void)
{
	int		curr_year = curr_tm.tm_year + 1900,
			curr_mon = curr_tm.tm_mon + 1;
	u_int		i;
	char		*path;
	char		oneline[DBRECORDSIZE + 1];
	struct stat	statbuf;
	struct ipac	*acp;
	struct tm	db_tm1, db_tm2;
	u_quad_t	db_value;

	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		acp->bcnt = 0ULL;
		if (process_mon == curr_mon && process_year == curr_year) {
			/* new day came, but the same month,
			   just add entry with 0 bytes for new day */
			if (append_db(acp, 0ULL) < 0)
				return -1;
			continue;
		}
		/* new month came */
		if (acp->fp != NULL) {
			/* close file for previous month */
			if (fclose(acp->fp) != 0)
				syslog(LOG_WARNING, "fclose(%s): %m", acp->filename);
			free(acp->filename);
		}
		if (asprintf(&path, "%s/%s/%d", acctdir, acp->rulename, curr_year) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			/* create directory ACCTDIR/<rule-name>/YYYY */
			if (mkdir_db(path) < 0)
				return -1;
		}
		free(path);

		if (asprintf(&path, "%s/%s/%d/%02d", acctdir, acp->rulename, curr_year, curr_mon) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			/* new month came, create new file and append record
			   with 0 bytes */
			if ( (acp->fp = create_db_file(path)) == NULL)
				return -1;
			acp->filename = path;
			if (append_db(acp, 0ULL) < 0)
				return -1;
			continue;
		}
		/* new month came but file for it already exists,
		   this is also hook for the first call */
		if (firstcall == 0)
			syslog(LOG_WARNING, "new month %d/%02d came, but file %s for it already exists",
			    curr_year, curr_mon, path);
		if ( (acp->fp = fopen(path, "r+")) == NULL) {
			syslog(LOG_ERR, "fopen(%s, \"r+\"): %m", path);
			return -1;
		}
		acp->filename = path;
		if (statbuf.st_size < DBRECORDSIZE) {
			syslog(LOG_ERR, "wrong format of file %s: too small file size", acp->filename);
			return -1;
		}
		if (fseek(acp->fp, -DBRECORDSIZE, SEEK_END) < 0) {
			syslog(LOG_ERR, "fseek(%s, -%d, SEEK_END): %m", acp->filename, DBRECORDSIZE);
			return -1;
		}
		if (fread(oneline, sizeof(char), DBRECORDSIZE, acp->fp) != DBRECORDSIZE) {
			syslog(LOG_ERR, "fread(%s): %m", acp->filename);
			return -1;
		}
		oneline[DBRECORDSIZE] = '\0';
		if (REGEXEC(db_entry, oneline) != 0) {
			syslog(LOG_ERR, "wrong format of file %s: can't recognize last record", acp->filename);
			return -1;
		}
		memcpy(&db_tm1, &curr_tm, sizeof db_tm1); /* set year, month, TZ, etc. */
		memcpy(&db_tm2, &curr_tm, sizeof db_tm2); /* set year, month, TZ, etc. */
		if (sscanf(oneline, "%d/%d:%d:%d-%d:%d:%d %qu", &db_tm1.tm_mday,
		    &db_tm1.tm_hour, &db_tm1.tm_min, &db_tm1.tm_sec,
		    &db_tm2.tm_hour, &db_tm2.tm_min, &db_tm2.tm_sec,
		    &db_value) != 8) {
			syslog(LOG_ERR, "%s: sscanf(%s, %%d/%%d:%%d:%%d-%%d:%%d:%%d %%qu): %m",
			    acp->filename, oneline);
			return -1;
		}
		db_tm2.tm_mday = db_tm1.tm_mday;
		if (checktm(&db_tm1) < 0 ||
		    (checktm(&db_tm2) < 0 && !(db_tm2.tm_hour == 24 && db_tm2.tm_min == 0 && db_tm2.tm_sec == 0)) ||
		    tmcmp(&db_tm1, &db_tm2) > 0) {
			syslog(LOG_ERR, "wrong format of file %s: wrong timestamp in last record", acp->filename);
			return -1;
		}
		if (tmcmp(&db_tm2, &curr_tm) > 0) {
			syslog(LOG_WARNING, "file %s: last record is more fresh than new one: append new record", acp->filename);
			if (append_db(acp, 0ULL) < 0)
				return 0;
			continue;
		}
		if (acp->append_db_time != 0) {
			/* "append_db_time" parameter is used */
			time_t	record_time;

			record_time = mktime(&db_tm1); /* XXX can be wrong value, because TZ has effect */
			if (curr_tm.tm_mday != db_tm1.tm_mday ||
			    record_time + acp->append_db_time <= curr_time) {
				/* new day came or was old record */
				if (append_db(acp, 0ULL) < 0)
					return -1;
			} else {
				/* record is present and is fresh,
				   store old value, but set newrec_time */
				acp->bcnt = db_value;
				acp->newrec_time = record_time + acp->append_db_time;
			}
		} else if (curr_tm.tm_mday != db_tm1.tm_mday) {
			/* new day */
			if (append_db(acp, 0ULL) < 0)
				return -1;
		} else
			/* record is present and fresh, store old byte counter */
			acp->bcnt = db_value;
	}
	process_mon = curr_tm.tm_mon + 1;
	process_year = curr_tm.tm_year + 1900;
	/* get time UTC of new day in seconds */
	newday_time = mk_time_exp_upto(curr_time, 'D');
	return 0;
}

int
run_ipac(void)
{
	int		ret;
	time_t		new_curr_time;
	struct tm	new_curr_tm;
	static sigset_t	sigmask, zeromask, pendmask;

	sigemptyset(&zeromask);
	sigemptyset(&sigmask);

	if (debug)
		sigaddset(&sigmask, SIGINT);
	sigaddset(&sigmask, SIGTERM);
	sigaddset(&sigmask, SIGHUP);
	sigaddset(&sigmask, SIGALRM);
	sigaddset(&sigmask, SIGCHLD);

	if (sigprocmask(SIG_BLOCK, &sigmask, (sigset_t *)0) < 0) {
		syslog(LOG_ERR, "sigprocmask(SIG_BLOCK): %m");
		return -1;
	}
	/*
	 * XXX: signals SIGTERM (and SIGINT), SIGHUP before function
	 *      sigprocmask() are not handled safe.
	 */
	syslog(LOG_INFO, "make IP accounting...");

	if (newday_came() < 0)
		return -1;

	for (;;) {
		if (use_ipfw)
			if (kipfw_read_table() < 0)
				return -1;
		if (use_ipfil)
			if (kipfil_read_table() < 0)
				return -1;

		time(&new_curr_time);
		localtime_r(&new_curr_time, &new_curr_tm);

		/* 
		   Check for problems with time and local date,
		   XXX: if TZ is changed outside, this fact isn't reflected
			on running process [is it bug of system?].
		*/
		if (new_curr_time < curr_time) {
			syslog(LOG_WARNING, "time UTC goes back? old time = %lds UTC, new time = %lds UTC (delta = -%s)",
			    curr_time, new_curr_time, show_time(curr_time - new_curr_time));
			time_changed_flag = 1;
		}
		if ((curr_tm.tm_zone && !new_curr_tm.tm_zone) ||
		    (!curr_tm.tm_zone && new_curr_tm.tm_zone) ||
		    (curr_tm.tm_zone && new_curr_tm.tm_zone && strcmp(curr_tm.tm_zone, new_curr_tm.tm_zone) != 0)) {
			syslog(LOG_WARNING, "time zone was changed, old TZ = %s, new TZ = %s",
			    curr_tm.tm_zone ? curr_tm.tm_zone : "???",
			    new_curr_tm.tm_zone ? new_curr_tm.tm_zone : "???");
			time_changed_flag = 1;
		}
		if (curr_tm.tm_gmtoff != new_curr_tm.tm_gmtoff) {
			syslog(LOG_WARNING, "offset from UTC was changed, old gmtoff = %lds, new gmtoff = %lds (delta = %s%ss)",
			    curr_tm.tm_gmtoff, new_curr_tm.tm_gmtoff,
			    curr_tm.tm_gmtoff > new_curr_tm.tm_gmtoff ? "" : "-",
			    curr_tm.tm_gmtoff > new_curr_tm.tm_gmtoff ? show_time(curr_tm.tm_gmtoff - new_curr_tm.tm_gmtoff) : show_time(new_curr_tm.tm_gmtoff - curr_tm.tm_gmtoff));
			time_changed_flag = 1;
		}
		if (curr_tm.tm_isdst != new_curr_tm.tm_isdst) {
			syslog(LOG_WARNING, "\"summer time in effect\" was changed, old isdst = %s, new isdst = %s",
			    curr_tm.tm_isdst ? "yes" : "no", new_curr_tm.tm_isdst ? "yes" : "no");
			time_changed_flag = 1;
		}
		if (tmcmp(&new_curr_tm, &curr_tm) < 0) {
			syslog(LOG_WARNING, "local time goes back? old localtime = %d.%02d.%02d/%02d:%02d:%02d (%s), new localtime = %d.%02d.%02d/%02d:%02d:%02d (%s)",
			    curr_tm.tm_year + 1900, curr_tm.tm_mon + 1, curr_tm.tm_mday, curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec,
			    curr_tm.tm_zone ? curr_tm.tm_zone : "???",
			    new_curr_tm.tm_year + 1900, new_curr_tm.tm_mon + 1, new_curr_tm.tm_mday, new_curr_tm.tm_hour, new_curr_tm.tm_min, new_curr_tm.tm_sec,
			    new_curr_tm.tm_zone ? new_curr_tm.tm_zone : "???");
			time_changed_flag = 1;
		}

		curr_time = new_curr_time;

		if (time_changed_flag) {
			syslog(LOG_WARNING, "unsuccessive date or time changes are detected");
			syslog(LOG_WARNING, "flush all counters and limits with old timestamp");
			/* 
			 * XXX store new accounting with old timestamp,
			 *     curr_tm isn't changed and is equal to old curr_tm.
			 */
			if (do_ipac() < 0)
				return -1;
			syslog(LOG_WARNING, "append new records for all rules");
			memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);
			firstcall = 1;
			process_mon = process_year = 0;
			if (newday_came() < 0)
				return -1;
			if (init_limits(0) < 0)
				return -1;
			time_changed_flag = 0;
			continue;
		}

		if (curr_time >= newday_time) {
			if (new_curr_tm.tm_mday != curr_tm.tm_mday) {
				newday_is_comming = 1;
				/* update record at the end of day with
				   second timestamp equal to 24:00:00 */
				curr_tm.tm_hour = 24;
				curr_tm.tm_min = curr_tm.tm_sec = 00;
				if (do_ipac() < 0)
					return -1;
				newday_is_comming = 0;
				memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);
				if (newday_came() < 0)
					return -1;
			} else
				syslog(LOG_WARNING, "new day didn't come, but was expected");
			memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);
			process_mon = curr_tm.tm_mon + 1;
			process_year = curr_tm.tm_year + 1900;
			/* get time UTC of new day in seconds */
			new_curr_tm.tm_hour = new_curr_tm.tm_min = new_curr_tm.tm_sec = 0;
			newday_time = mktime(&new_curr_tm) + DAY; 
		} else 
			memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);

		if (do_ipac() < 0)
			return -1;
		firstcall = 0;
		if (goingdown_flag)
			return 0;
		if (reconfigure_flag) {
			ret = do_reconfigure();
			reconfigure_flag = 0;
			if (ret == 0) {
				/* configuration file was successfuly parsed */
				if (newday_came() < 0)
					return -1;
				continue;
			}
			if (ret == -1)
				/* config file was successful parsed,
				   but kipfw_init() or kipfil_init() failed */
				return -1;
		}
		if (sleep_seconds + curr_time > newday_time)
			sleep_seconds = newday_time - curr_time;

		if (debug_time > 0)
			syslog(LOG_INFO, "run_ipac: sleep_seconds = %u sec.", sleep_seconds);

		if (sleep_seconds != 0) {
			alarm(sleep_seconds);	/* set alarm */
			while (wakeup_flag == 0)
				sigsuspend(&zeromask);
			alarm(0);		/* release alarm */
		} else {
			/* should not set alarm, also signals SIGINT, SIGTERM, SIGHUP
			   and SIGCHLD should be tested */
			if (sigpending(&pendmask) < 0) {
				syslog(LOG_ERR, "sigpending: %m");
				return -1;
			}
			if (sigismember(&pendmask, SIGINT) ||
			    sigismember(&pendmask, SIGTERM) ||
			    sigismember(&pendmask, SIGHUP) ||
			    sigismember(&pendmask, SIGCHLD))
				sigsuspend(&zeromask);	/* should be interrupted immediately */
		}
		wakeup_flag = 0;
	}
}

struct ipac *
ipac_new(void)
{
	u_int		i;
	struct ipac	*acp;

	if ( (acp = realloc(ipac, (nipac + 1) * sizeof *acp)) == NULL) {
		syslog(LOG_ERR, "realloc: %m");
		return NULL;
	}
	ipac = acp;
	acp += nipac;
	++nipac;
	acp->filename = NULL;
	acp->fp = NULL;
	acp->bcnt_sub = 0;
	acp->rulename = NULL;
	acp->filename = NULL;
	acp->ipfwac = NULL;
	acp->ipfilac = NULL;
	acp->limit = NULL;
	acp->rc[0].cmd = acp->rc[0].cmd_if_limit = acp->rc[0].cmd_if_nolimit = NULL;
	acp->rc[1].cmd = acp->rc[1].cmd_if_limit = acp->rc[1].cmd_if_nolimit = NULL;
	acp->acl = NULL;
	acp->info = NULL;
	acp->maxchunk = 0;
	acp->wakeup = 0;
	for (i = 0; i < 7; ++i)
		acp->worktime[i].interval = NULL;
	return acp;
}

static int
check_db_dir_ownership(const char *path, struct stat *statbuf)
{
	if (!S_ISDIR(statbuf->st_mode)) {
		syslog(LOG_ERR, "%s should be directory", path);
		return -1;
	}
	if (statbuf->st_uid != db_uid || statbuf->st_gid != db_gid) {
		syslog(LOG_WARNING, "directory %s is owned by UID %u GID %u",
		    path, statbuf->st_uid, statbuf->st_gid);
		syslog(LOG_WARNING, "changing ownership of directory %s to UID %lu GID %lu", path, db_uid, db_gid);
		if (chown_db(path) < 0)
			return -1;
	}
	if (PERMMASK(statbuf->st_mode) != db_dirperm) {
		syslog(LOG_WARNING, "directory %s has permissions 0%03o",
		    path, PERMMASK(statbuf->st_mode));
		syslog(LOG_WARNING, "changing permissions for directory %s to 0%03o",
		    path, db_dirperm);
		if (chmod(path, db_dirperm) < 0) {
			syslog(LOG_ERR, "chmod(%s, 0%03o): %m", path, db_dirperm);
			return -1;
		}
	}
	return 0;
}

static int
check_db_file_ownership(const char *path, struct stat *statbuf)
{
	if (!S_ISREG(statbuf->st_mode)) {
		syslog(LOG_ERR, "%s should be regular file", path);
		return -1;
	}
	if (statbuf->st_uid != db_uid || statbuf->st_gid != db_gid) {
		syslog(LOG_WARNING, "file %s is owned by UID %u GID %u",
		    path, statbuf->st_uid, statbuf->st_gid);
		syslog(LOG_WARNING, "changing owner of file %s to UID %lu GID %lu",
		    path, db_uid, db_gid);
		if (chown_db(path) < 0)
			return -1;
	}
	if (PERMMASK(statbuf->st_mode) != db_fileperm) {
		syslog(LOG_WARNING, "file %s has permissions 0%03o",
		    path, PERMMASK(statbuf->st_mode));
		syslog(LOG_WARNING, "changing permissions for file %s to 0%03o",
		    path, db_fileperm);
		if (chmod(path, db_fileperm) < 0) {
			syslog(LOG_ERR, "chmod(%s, 0%03o): %m", path, db_fileperm);
			return -1;
		}
	}
	return 0;
}

static int
mkdir_db(const char *path)
{
	if (mkdir(path, db_dirperm) < 0) {
		syslog(LOG_ERR, "mkdir(%s, 0%03o): %m", path, db_dirperm);
		return -1;
	}
	return chown_db(path);
}

static FILE *
create_db_file(const char *path)
{
	int	fd;
	FILE	*fp;

	if ( (fd = open(path, O_RDWR|O_CREAT, db_fileperm)) < 0) {
		syslog(LOG_ERR, "open(%s, O_RDWR|O_CREAT, 0%03o): %m", path, db_fileperm);
		return NULL;
	}
	if (chown_db(path) < 0)
		return NULL;
	/* some file can be already exists, truncate it */
	if (ftruncate(fd, (off_t)0) < 0) {
		syslog(LOG_ERR, "ftruncate(FD %d, 0), truncate %s: %m", fd, path);
		return NULL;
	}
	if ( (fp = fdopen(fd, "r+")) == NULL)
		syslog(LOG_ERR, "fdopen(FD %d, \"r+\"), reopen %s: %m", fd, path);
	return fp;
}

int
init_db_rules(void)
{
	char		*path, *acl;
	u_int		i, j;
	FILE		*fp;
	struct ipac	*acp;
	struct limit	*lp;
	struct stat	statbuf;

	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		/* check directory for rule */
		if (asprintf(&path, "%s/%s", acctdir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			if (errno != ENOENT)
				syslog(LOG_WARNING, "stat(%s): %m", path);
			syslog(LOG_INFO, "creating rule directory %s", path);
			if (mkdir_db(path) < 0)
				return -1;
		}
		free(path);

		/* check directory for limits */
		if (asprintf(&path, "%s/%s/" LIMITSDIR, acctdir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			if (errno != ENOENT)
				syslog(LOG_WARNING, "stat(%s): %m", path);
			if (mkdir_db(path) < 0)
				return -1;
		}
		free(path);

		/* create/overwrite acl file */
		if (asprintf(&path, "%s/%s/" ACLFILE, acctdir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if ( (acl = acp->acl ? acp->acl : acl_global) != NULL) {
			if ( (fp = create_db_file(path)) == NULL)
				return -1;
			if (fprintf(fp, "%s", acl) != strlen(acl)) {
				syslog(LOG_ERR, "fprintf(%s): %m", path);
				return -1;
			}
			if (fclose(fp) != 0) {
				syslog(LOG_ERR, "fclose(%s): %m", path);
				return -1;
			}
			free(acp->acl);
			acp->acl = NULL;
		} else {
			if (unlink(path) < 0)
				if (errno != ENOENT) {
					syslog(LOG_ERR, "unlink(%s): %m", path);
					return -1;
				}
		}
		free(path);

		/* create/owerwrite info file */
		if (asprintf(&path, "%s/%s/" INFOFILE, acctdir, acp->rulename) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (acp->info) {
			if ( (fp = create_db_file(path)) == NULL)
				return -1;
			if (fprintf(fp, "%s", acp->info) != strlen(acp->info)) {
				syslog(LOG_ERR, "fprintf(%s): %m", path);
				return -1;
			}
			if (fclose(fp) != 0) {
				syslog(LOG_ERR, "fclose(%s): %m", path);
				return -1;
			}
			free(acp->info);
			acp->info = NULL;
		} else {
			if (unlink(path) < 0)
				if (errno != ENOENT) {
					syslog(LOG_ERR, "unlink(%s): %m", path);
					return -1;
				}
		}
		free(path);

		/* create directories for limits */
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++j, ++lp) {
			/* create directory for limit */
			if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s", acctdir,
			    acp->rulename, lp->limitname) < 0) {
				syslog(LOG_ERR, "asprintf: %m");
				return -1;
			}
			if (stat(path, &statbuf) < 0) {
				if (errno != ENOENT)
					syslog(LOG_ERR, "stat(%s): %m", path);
				syslog(LOG_INFO, "creating directory for rule %s limit %s",
				    acp->rulename, lp->limitname);
				if (mkdir_db(path) < 0)
					return -1;
			}
			free(path);

			/* create or open file for limit accounting */
			if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s/" LIMITFILE,
			    acctdir, acp->rulename, lp->limitname) < 0) {
				syslog(LOG_ERR, "asprintf: %m");
				return -1;
			}
			if (stat(path, &statbuf) < 0) {
				if (errno != ENOENT)
					syslog(LOG_WARNING, "stat(%s): %m", path);
				if ( (lp->fp = create_db_file(path)) == NULL)
					return -1;
			} else {
				if (statbuf.st_size == 0)
					syslog(LOG_WARNING, "file %s is empty, it will be fixed.", path);
				if ( (lp->fp = fopen(path, "r+")) == NULL) {
					syslog(LOG_ERR, "fopen(%s, \"r+\"): %m", path);
					return -1;
				}
			}
			lp->filename = path;

			/* create info file for limit */
			if (asprintf(&path, "%s/%s/" LIMITSDIR "/%s/" INFOFILE,
			    acctdir, acp->rulename, lp->limitname) < 0) {
				syslog(LOG_ERR, "asprintf: %m");
				return -1;
			}
			if (lp->info) {
				if ( (fp = create_db_file(path)) == NULL)
					return -1;
				if (fprintf(fp, "%s", lp->info) != strlen(lp->info)) {
					syslog(LOG_ERR, "fprintf(%s): %m", path);
					return -1;
				}
				if (fclose(fp) != 0) {
					syslog(LOG_ERR, "fclose(%s): %m", path);
					return -1;
				}
				free(lp->info);
				lp->info = NULL;
			} else {
				if (unlink(path) < 0)
					if (errno != ENOENT) {
						syslog(LOG_ERR, "unlink(%s): %m", path);
						return -1;
					}
			}
			free(path);
		}
	}
	if (acl_global) {
		free(acl_global);
		acl_global = NULL;
	}
	return 0;
}

static int
append_limit_ts(struct limit *lp, char type, struct tm *tm_ptr)
{
	if (fseek(lp->fp, 0, SEEK_END) < 0) {
		syslog(LOG_ERR, "fseek(%s, 0L, SEEK_END): %m", lp->limitname);
		return -1;
	}
	if (fprintf(lp->fp, "%c %d.%02d.%02d/%02d:%02d:%02d\n", type,
	    tm_ptr->tm_year + 1900, tm_ptr->tm_mon + 1, tm_ptr->tm_mday,
	    tm_ptr->tm_hour, tm_ptr->tm_min, tm_ptr->tm_sec) < LIMIT_TIMESTAMP_SIZE) {
		syslog(LOG_ERR, "fprintf(%s): %m", lp->limitname);
		return -1;
	}
	if (fflush(lp->fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", lp->filename);
		return -1;
	}
	return 0;
}

static int
check_db_deep(char *start)
{
	char		*path;
	DIR		*dp;
	struct stat	statbuf;
	struct dirent	*dirp;

	if ( (dp = opendir(start)) == NULL) {
		syslog(LOG_ERR, "opendir(%s): %m", start);
		return -1;
	}
	while ( (dirp = readdir(dp)) != 0) {
		if (strcmp(dirp->d_name, ".") == 0 || strcmp(dirp->d_name, "..") == 0)
			continue;
		if (asprintf(&path, "%s/%s", start, dirp->d_name) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		if (stat(path, &statbuf) < 0) {
			syslog(LOG_ERR, "stat(%s): %m", path);
			return -1;
		}
		if (S_ISREG(statbuf.st_mode)) {
			if (check_db_file_ownership(path, &statbuf) < 0)
				return -1;
		} else if (S_ISDIR(statbuf.st_mode)) {
			if (check_db_dir_ownership(path, &statbuf) < 0)
				return -1;
			if (check_db_deep(path) < 0)
				return -1;
		} else {
			syslog(LOG_ERR, "%s must be a regular file or directory", path);
			return -1;
		}
		free(path);
	}
	if (closedir(dp) < 0) {
		syslog(LOG_ERR, "closedir(%s): %m", start);
		return -1;
	}
	return 0;
}

static int 
check_db_ownership(void)
{
	struct stat	statbuf;

	if (stat(acctdir, &statbuf) < 0) {
		syslog(LOG_ERR, "stat(%s): %m", acctdir);
		return -1;
	}
	if (statbuf.st_uid != db_uid || statbuf.st_gid != db_gid ||
	    PERMMASK(statbuf.st_mode) != db_dirperm) {
		syslog(LOG_WARNING, "directory %s has wrong permissions or is onwed wrong", acctdir);
		syslog(LOG_INFO, "checking all data base...");
		if (check_db_dir_ownership(acctdir, &statbuf) < 0)
			return -1;
		if (check_db_deep(acctdir) < 0)
			return -1;
	}
	return 0;
}

int
init_db(void)
{
	struct stat	statbuf;

	if (stat(acctdir, &statbuf) < 0) {
		if (errno != ENOENT)
			syslog(LOG_WARNING, "stat(%s): %m", acctdir);
		syslog(LOG_INFO, "creating main directory %s", acctdir);
		if (mkdir_db(acctdir) < 0)
			return -1;
	}

	if (check_db_ownership() < 0)
		return -1;

	if (init_db_rules() < 0)
		return -1;

	time(&curr_time);
	localtime_r(&curr_time, &curr_tm);
	return init_limits(1);
}

int
build_db_regexes(void)
{
#define	REGCOMP(x)\
	if (regcomp(&reg_ ## x, pat_ ## x, REG_EXTENDED) != 0) {	\
		syslog(LOG_ERR, "regcomp(" #x ")");			\
		return -1;						\
	}

	REGCOMP(db_entry);
	REGCOMP(limit_line1);
	REGCOMP(limit_timestamp);
	return 0;
}

void
end_work(int signo)
{
	syslog(LOG_INFO, "caught signal %d '%s', shutdowning...", signo, sys_signame[signo]);
	goingdown_flag = wakeup_flag = 1;
}

void
sig_alrm(int signo)
{
	alarm_flag = wakeup_flag = 1;
}

void
reconfigure(int signo)
{
	syslog(LOG_WARNING, "caught signal %d '%s', reconfiguring...", signo, sys_signame[signo]);
	reconfigure_flag = wakeup_flag = 1;
}

int
do_reconfigure(void)
{
	struct ipac	*ipac_bkp = ipac;
	u_int		nipac_bkp = nipac;
	struct commands	shutdown_global_bkp = shutdown_global;
	int		use_ipfw_bkp = use_ipfw,
			use_ipfil_bkp = use_ipfil;
	uid_t		db_uid_bkp = db_uid;
	gid_t		db_gid_bkp = db_gid;

	syslog(LOG_INFO, "rereading configuration file...");
	ipac = NULL;
	nipac = 0;
	use_ipfw = use_ipfil = 0;

	if (parse_config(RECONFIG_PARSING) < 0) {
		syslog(LOG_WARNING, "continue using previous configuration");
		if (free_ipac_list(ipac, nipac) < 0)
			return -1;
		ipac = ipac_bkp;
		nipac = nipac_bkp;
		use_ipfw = use_ipfw_bkp;
		use_ipfil = use_ipfil_bkp;
		db_uid = db_uid_bkp;
		db_gid = db_gid_bkp;
		shutdown_global = shutdown_global_bkp;
		return -2;
	}

	if (use_ipfw_bkp == 1 && use_ipfw == 0)
		kipfw_close();
	else if (use_ipfw_bkp == 0 && use_ipfw == 1)
		if (kipfw_init() < 0)
			return -1;

	if (use_ipfil_bkp == 1 && use_ipfil == 0)
		kipfil_close();
	else if (use_ipfil_bkp == 0 && use_ipfil == 1)
		if (kipfil_init() < 0)
			return -1;

	syslog(LOG_INFO, "use new configuration");
	if (free_ipac_list(ipac_bkp, nipac_bkp) < 0)
		return -1;
	free_cmd_list(shutdown_global.cmd, shutdown_global.ncmd);
	process_mon = process_year = 0;
	firstcall = 1;
	return init_db();
}

int
free_ipac_list(struct ipac *ipac_head, u_int n)
{
	struct ipac	*acp;
	struct ipfilac	*ipfilacp;
	struct limit	*lp;
	u_int		i, j;

	for (acp = ipac_head, i = 0; i < n; ++acp, ++i) {
		if (acp->rulename)
			free(acp->rulename);
		if (acp->fp) {
			if (fclose(acp->fp) != 0) {
				syslog(LOG_ERR, "close(%s): %m", acp->filename);
				return -1;
			}
			free(acp->filename);
		}
		if (acp->ipfwac)
			free(acp->ipfwac);
		if (acp->ipfilac) {
			for (ipfilacp = acp->ipfilac, j = 0; j < acp->nipfilac; ++ipfilacp, ++j)
				if (ipfilacp->list)
					free(ipfilacp->list);
			free(acp->ipfilac);
		}
		if (acp->limit) {
			for (lp = acp->limit, j = 0; j < acp->nlimit; ++lp, ++j) {
				if (lp->limitname)
					free(lp->limitname);
				if (lp->fp) {
					if (fclose(lp->fp) != 0) {
						syslog(LOG_ERR, "fclose(%s): %m", lp->filename);
						return -1;
					}
					free(lp->filename);
				}
				free_cmd_list(lp->cmd, lp->ncmd);
				free_cmd_list(lp->expire.cmd, lp->expire.ncmd);
				free_commands(&lp->rc[0]);
				free_commands(&lp->rc[1]);
			}
			free(acp->limit);
		}
		free_commands(&acp->rc[0]);
		free_commands(&acp->rc[1]);
	}
	free(ipac_head);
	return 0;
}

int
update_limits(struct ipac *acp)
{
	u_int		i;
	struct limit	*lp;

	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i)
		if (lp->is_changed) {
			if (fseek(lp->fp, 0L, SEEK_SET) < 0) {
				syslog(LOG_ERR, "fseek(%s, 0, SEEK_SET): %m", lp->filename);
				return -1;
			}
			if (fprintf(lp->fp, "%020qu", lp->bcnt) != LIMIT_COUNTER_SIZE) {
				syslog(LOG_ERR, "fprintf(%s, \"%%020qu\"): %m", lp->filename);
				return -1;
			}
			if (fflush(lp->fp) != 0) {
				syslog(LOG_ERR, "fflush(%s): %m", lp->filename);
				return -1;
			}
			lp->is_changed = 0;
		}
	return 0;
}

void
add_chunk_to_rule_limits(struct ipac *acp, u_quad_t chunk)
{
	u_int		i;
	u_quad_t	chunk_tmp;
	struct limit	*lp;

	if (chunk == 0)
		return;
	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i) {
		if (lp->status & LIMIT_IS_REACHED)
			continue;
		if (lp->bcnt_sub >= chunk)
			lp->bcnt_sub -= chunk;
		else {
			chunk_tmp = chunk - lp->bcnt_sub;
			lp->bcnt_sub = 0;
			if (lp->bcnt < UQUAD_MAX - chunk_tmp)
				lp->bcnt += chunk_tmp;
			else {
				lp->bcnt = UQUAD_MAX;
				syslog(LOG_WARNING, "rule %s limit %s: byte counter overflowed",
				    acp->rulename, lp->limitname);
			}
			lp->is_changed = 1;
		}
	}
}

void
sub_chunk_from_rule_limits(struct ipac *acp, u_quad_t chunk)
{
	u_int		i;
	struct limit	*lp;

	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i) {
		if (lp->status & LIMIT_IS_REACHED)
			continue;
		if (lp->bcnt >= chunk)
			lp->bcnt -= chunk;
		else {
			if (lp->bcnt_sub < UQUAD_MAX - (chunk - lp->bcnt))
				lp->bcnt_sub += chunk - lp->bcnt;
			else {
				syslog(LOG_WARNING, "rule %s limit %s: wrong configuration for some ipfw or ipfil rule",
				    acp->rulename, lp->limitname);
				lp->bcnt_sub = UQUAD_MAX;
			}
			lp->bcnt = 0;
		}
		lp->is_changed = 1;
	}
}

int
chown_db(const char *path)
{
	if (chown(path, db_uid, db_gid) < 0) {
		syslog(LOG_ERR, "chown(%s, %lu, %lu): %m", path, db_uid, db_gid);
		return -1;
	}
	return 0;
}

int
run_rules_rc(int x)
{
	u_int		i, j;
	int		limit_is_reached;
	char		*msg = x == 0 ? "startup" : "shutdown";
	struct ipac	*acp;
	struct limit	*lp;
	
	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		if (acp->rc[x].cmd) {
			syslog(LOG_INFO, "rule %s (%s): run commands", acp->rulename, msg);
			if (exec_cmd_list(acp->rc[x].cmd, acp->rc[x].ncmd,
			    "rule %s (%s)", acp->rulename, msg) < 0)
				return -1;
		}
		limit_is_reached = 0;
		if (acp->rc[x].cmd_if_limit || acp->rc[x].cmd_if_nolimit) {
			limit_is_reached = 0;
			for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j)
				if (lp->bcnt >= lp->byte_limit) {
					limit_is_reached = 1;
					break;
				}
			if (limit_is_reached && acp->rc[x].cmd_if_limit) {
				syslog(LOG_INFO, "rule %s (%s, if_limit): run commands", acp->rulename, msg);
				if (exec_cmd_list(acp->rc[x].cmd_if_limit, acp->rc[x].ncmd_if_limit,
				    "rule %s (%s, if_limit)", acp->rulename, msg) < 0)
				return -1;
			} else if (!limit_is_reached && acp->rc[x].cmd_if_nolimit) {
				syslog(LOG_INFO, "rule %s (%s, if_nolimit): run commands", acp->rulename, msg);
				if (exec_cmd_list(ipac->rc[x].cmd_if_nolimit, acp->rc[x].ncmd_if_nolimit,
				    "rule %s (%s, if_nolimit)", acp->rulename, msg) < 0)
					return -1;
			}
		}
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j) {
			if (lp->rc[x].cmd) {
				syslog(LOG_INFO, "rule %s limit %s (%s): run commands", acp->rulename, lp->limitname, msg);
				if (exec_cmd_list(lp->rc[x].cmd, lp->rc[x].ncmd,
				    "rule %s limit %s (%s)", acp->rulename, lp->limitname, msg) < 0)
					return -1;
			}
			if (lp->status & LIMIT_IS_REACHED) {
				if (lp->rc[x].cmd_if_limit) {
					syslog(LOG_INFO, "rule %s limit %s (%s, if_limit): run commands",
					    acp->rulename, lp->limitname, msg);
					if (exec_cmd_list(lp->rc[x].cmd_if_limit, lp->rc[x].ncmd_if_limit,
					    "rule %s limit %s (%s, if_limit)", acp->rulename, lp->limitname, msg) < 0)
						return -1;
				}
			} else if (lp->rc[x].cmd_if_nolimit) {
				syslog(LOG_INFO, "rule %s limit %s (%s, if_nolimit): run commands",
				    ipac->rulename, lp->limitname, msg);
				if (exec_cmd_list(lp->rc[x].cmd_if_nolimit, lp->rc[x].ncmd_if_nolimit,
				    "rule %s limit %s (%s, if_nolimit)", acp->rulename, lp->limitname, msg) < 0)
					return -1;
			}
		}
	}
	return 0;
}

time_t
check_rule_limits(struct ipac *acp, time_t *wakeup)
{
	u_int		i;
	struct limit	*lp;

	*wakeup = curr_time + DAY; /* just initial value */
	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i) {
		if (lp->status & LIMIT_IS_REACHED) {
			/* limit has been already reached */
			if (lp->status & LIMIT_EXEC)
				continue; /* some commands are run in background */
			if (lp->expire.time.upto != UPTO_NOTSET) {
				/* "expire" section is used */
				if (!(lp->status & LIMIT_EXPIRE_CHECK)) {
					if (append_limit_e_ts(lp) < 0)
						return -1;
					lp->status |= LIMIT_EXPIRE_CHECK;
					if (*wakeup > lp->expire_time)
						*wakeup = lp->expire_time;
				} else if (lp->expire_time <= curr_time) {
					/* it's time to expire limit */
					if (lp->expire.cmd != NULL && !(lp->status & LIMIT_EXPIRE_EXEC)) {
						/* "exec" parameter in "expire" section is used */
						if (debug_limit > 0) {
							syslog(LOG_INFO, "rule %s limit %s: limit expired",
							    acp->rulename, lp->limitname);
							syslog(LOG_INFO, "rule %s limit %s (expire, exec): run commands",
							    acp->rulename, lp->limitname);
						}
						if ( (lp->pid = exec_cmd_list_bg(lp->expire.cmd, lp->expire.ncmd, "rule %s limit %s (expire, exec)",
						    acp->rulename, lp->limitname)) < 0)
							return -1;
						if (append_limit_ts(lp, LIMIT_EXECUTED, &curr_tm) < 0)
							return -1;
						lp->status |= LIMIT_EXPIRE_EXEC | LIMIT_EXEC;
					} else if (lp->expire.cmd == NULL ||
						   (lp->expire.cmd != NULL && (lp->status & LIMIT_EXPIRE_EXEC))) {
						/* "exec" parameter in "expire" section isn't used,
						   or is used, but commands already have been executed */
						if (lp->expire.cmd == NULL && debug_limit > 0)
							syslog(LOG_INFO, "rule %s limit %s: limit expired",
							    acp->rulename, lp->limitname);
						if (restart_limit(acp, lp) < 0)
							return -1;
						if (lp->zero_time_param.upto != UPTO_NOTSET)
							if (*wakeup > lp->zero_time)
								*wakeup = lp->zero_time;
						lp->status = 0;
					}
				} else {
					/* just check time to expire */
					if (*wakeup > lp->expire_time)
						*wakeup = lp->expire_time;
				}
			}
		} else {
			/* limit hasn't been reached before */
			if (lp->bcnt >= lp->byte_limit) {
				/* limit has just been reached */
				if (debug_limit > 0)
					syslog(LOG_INFO, "rule %s limit %s: limit reached, bcnt = %s",
					    acp->rulename, lp->limitname, show_size(lp->bcnt));
				lp->status = LIMIT_IS_REACHED;
				lp->reach_time = curr_time;
				if (append_limit_ts(lp, LIMIT_REACHED, &curr_tm) < 0)
					return -1;
				if (lp->cmd != NULL) {
					/* "exec" parameter in "limit" section is used */
					if (debug_limit)
						syslog(LOG_INFO, "rule %s limit %s (reach): run commands", 
						    acp->rulename, lp->limitname);
					if ( (lp->pid = exec_cmd_list_bg(lp->cmd, lp->ncmd, "rule %s limit %s (reach)",
					    acp->rulename, lp->limitname)) < 0)
						return -1;
					if (append_limit_ts(lp, LIMIT_EXECUTED, &curr_tm) < 0)
						return -1;
					lp->status |= LIMIT_EXEC;
				} else if (lp->expire.time.upto != UPTO_NOTSET) {
					if (append_limit_e_ts(lp) < 0)
						return -1;
					lp->status |= LIMIT_EXPIRE_CHECK;
					if (*wakeup > lp->expire_time)
						*wakeup = lp->expire_time;
				}
			} else if (lp->zero_time_param.upto != UPTO_NOTSET) {
				/* "zero_time" parameter in "limit" section is used */
				if (lp->zero_time <= curr_time) {
					/* it is time to restart limit */
					if (debug_limit)
						syslog(LOG_INFO, "rule %s limit %s: limit zeroed",
						    acp->rulename, lp->limitname);
					if (restart_limit(acp, lp) < 0)
						return -1;
				}
				if (*wakeup > lp->zero_time)
					*wakeup = lp->zero_time;
			}
		}
	}
	return 0;
}

time_t
mk_time_exp_upto(time_t start_time, char upto)
{
	time_t		result_time = start_time;
	struct tm	tmp_tm;

	if (upto == ' ')
		return start_time;

	localtime_r(&result_time, &tmp_tm);

	if (upto == 'M') {
		/* up to the end of month */
		if (tmp_tm.tm_mon == 11) {
			tmp_tm.tm_mon = 0;
			++tmp_tm.tm_year;
		} else
			++tmp_tm.tm_mon;
		tmp_tm.tm_mday = 1;
		tmp_tm.tm_hour = tmp_tm.tm_min = tmp_tm.tm_sec = 0;
		return mktime(&tmp_tm);
	}
	/* up to the end of minute */
	result_time += MINUTE - tmp_tm.tm_sec;
	++tmp_tm.tm_min;
	if (upto == 'm')
		return result_time;
	
	/* up to the end of hour */
	result_time += MINUTE * (60 - tmp_tm.tm_min);
	++tmp_tm.tm_hour;
	if (upto == 'h')
		return result_time;
	
	/* up to the end of day */
	result_time += HOUR * (24 - tmp_tm.tm_hour);
	++tmp_tm.tm_wday;
	if (upto == 'D')
		return result_time;

	/* up to the end of week */
	if (tmp_tm.tm_wday != 1)
		result_time += DAY * (8 - tmp_tm.tm_wday);
	return result_time;	/* upto == 'W' */
}

static time_t
mk_time_exp(time_t start_time, struct time_exp *texp)
{
	time_t		result_time, tmp_time;
	u_int		monthes, years;
	struct tm	result_tm, tmp1_tm;

	result_time = start_time;
	if (texp->side == 0)
		result_time = mk_time_exp_upto(start_time, texp->upto);
	result_time += texp->seconds;
	if (texp->monthes != 0) {
		localtime_r(&result_time, &result_tm);
		if (texp->monthes != 0) {
			if (result_tm.tm_mon + texp->monthes > 11) {
				monthes = texp->monthes - (11 - result_tm.tm_mon);
				years = monthes / 12;
				result_tm.tm_year += 1 + years;
				result_tm.tm_mon = monthes - 12 * years - 1;
			} else
				result_tm.tm_mon += texp->monthes;
			if (result_tm.tm_mday > 28) {
				memcpy(&tmp1_tm, &result_tm, sizeof result_tm);
				if (tmp1_tm.tm_mon == 11) {
					tmp1_tm.tm_mon = 0;
					++tmp1_tm.tm_year;
				} else
					++tmp1_tm.tm_mon;
				tmp1_tm.tm_mday = 1;
				tmp_time = mktime(&tmp1_tm);
				tmp_time -= DAY;
				localtime_r(&tmp_time, &tmp1_tm);
				if (result_tm.tm_mday > tmp1_tm.tm_mday)
					result_tm.tm_mday = tmp1_tm.tm_mday;
			}
		}
		result_time = mktime(&result_tm);
	}
	return texp->side == 0 ? result_time : mk_time_exp_upto(result_time, texp->upto);
}


int
restart_limit(struct ipac *acp, struct limit *lp)
{
	if (debug_limit > 0)
		syslog(LOG_INFO, "rule %s limit %s: restart limit",
		    acp->rulename, lp->limitname);
	if (Ftruncate_Seek(lp->fp, 0L, lp->filename) < 0)
		return -1;
	if (fprintf(lp->fp, "%020qu %020qu\n", 0ULL, lp->byte_limit) != LIMIT_LINE1_SIZE) {
		syslog(LOG_ERR, "fprintf(%s, \"%%020qu %%020qu\\n\"): %m", lp->filename);
		return -1;
	}
	if (append_limit_ts(lp, LIMIT_STARTED, &curr_tm) < 0)
		return -1;
	lp->start_time = curr_time;
	lp->bcnt = 0;
	return append_limit_z_ts(lp);
}

static void
log_limit_fix(struct ipac *acp, struct limit *lp)
{
	syslog(LOG_WARNING, "rule %s limit %s: errors fixed",
	    acp->rulename, lp->limitname);
}

/*
 * Use curr_tm and curr_time to determine if timestamps for limit are correct.
 * Limits files were opened in init_db_rules() function.
 */
int
init_limits(int mode)
{
	char		*buf = NULL;
	size_t		bufsize;
	u_int		i, j, k;
	int		len;
	u_quad_t	old_size;
	struct ipac	*acp;
	struct limit	*lp;
	struct timestamp {
		short		ts_is_set;
		long		ts_offset;
		struct tm	ts_tm;
	}		ts_start, ts_zero, ts_reach, ts_expire, ts_exec1,
			ts_exec2, *ts_ptr;

	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++j, ++lp) {
			len = readline(&buf, &bufsize, lp->fp, lp->filename);
			if (len == 0 && feof(lp->fp) != 0) {
				/* file has been just created or it is empty */
				if (restart_limit(acp, lp) < 0)
					return -1;
				continue;
			}
			if (len == 0) {
				/* XXX readline() can't return 0 without EOF */
				syslog(LOG_ERR, "read 0 bytes from file %s (first line)", lp->filename);
				return -1;
			}
			if (len < 0) {
				syslog(LOG_ERR, "can't read first line from file %s", lp->filename);
				return -1;
			}
			if (REGEXEC(limit_line1, buf) != 0) {
				syslog(LOG_ERR, "%s: can't recognize first line", lp->filename);
				return -1;
			}
			if (sscanf(buf, "%qu %qu", &lp->bcnt, &old_size) != 2) {
				syslog(LOG_ERR, "%s: sscanf(%s, \"%%qu %%qu\"): %m", lp->filename, buf);
				return -1;
			}
			if (lp->byte_limit != old_size && debug_limit > 0)
				syslog(LOG_WARNING, "rule %s limit %s: parameter \"size\" was changed, old size = %s new size = %s",
				    acp->rulename, lp->limitname, show_size(old_size), show_size2(lp->byte_limit));

			ts_start.ts_is_set = ts_zero.ts_is_set = ts_reach.ts_is_set = 
			    ts_expire.ts_is_set = ts_exec1.ts_is_set = ts_exec2.ts_is_set = 0;
			for (k = 0; k < 7; ++k) {
				len = readline(&buf, &bufsize, lp->fp, lp->filename);
				if (len == 0 && feof(lp->fp) != 0) {
					if (k == 0) {
						syslog(LOG_ERR, "%s: too small file size", lp->filename);
						log_limit_fix(acp, lp);
						if (restart_limit(acp, lp) < 0)
							return -1;
						goto next_limit;
					}
					break;
				}
				if (len > 0 && k == 6) {
					syslog(LOG_ERR, "%s: too big file size (more than 6 timestamps)", lp->filename);
					return -1;
				}
				if (len == 0) {
					/* XXX readline() can't return 0 without EOF */
					syslog(LOG_ERR, "%s: read 0 bytes (%d line)", lp->filename, k + 2);
					return -1;
				}
				if (len < 0) {
					syslog(LOG_ERR, "%s: can't read %d line from", lp->filename, k + 2);
					return -1;
				}
				if (REGEXEC(limit_timestamp, buf) != 0) {
					syslog(LOG_ERR, "%s: can't recognize %d line", lp->filename, k + 2);
					return -1;
				}
				ts_ptr = NULL;
				switch (*buf) {
				case LIMIT_STARTED:
					if (k != 0)
						break;
					if (!ts_start.ts_is_set)
						ts_ptr = &ts_start;
					break;
				case LIMIT_ZEROED:
					if (k != 1)
						break;
					if (!ts_zero.ts_is_set)
						ts_ptr = &ts_zero;
					break;
				case LIMIT_REACHED:
					if ((ts_zero.ts_is_set && k != 2) ||
					    (!ts_zero.ts_is_set && k != 1))
						break;
					if (!ts_reach.ts_is_set)
						ts_ptr = &ts_reach;
					break;
				case LIMIT_EXECUTED:
					if (k == 2 || k == 3) {
						if (!ts_exec1.ts_is_set)
							ts_ptr = &ts_exec1;
					} else if (k == 4 || k == 5)
						if (!ts_exec2.ts_is_set)
							ts_ptr = &ts_exec2;
					break;
				case LIMIT_EXPIRED:
					if ((!ts_zero.ts_is_set && !ts_exec1.ts_is_set && k != 2) ||
					    (!ts_zero.ts_is_set && ts_exec1.ts_is_set && k != 3) ||
					    (ts_zero.ts_is_set && !ts_exec1.ts_is_set && k != 3) ||
					    (ts_zero.ts_is_set && ts_exec1.ts_is_set && k != 4))
						break;
					if (!ts_expire.ts_is_set)
						ts_ptr = &ts_expire;
					break;
				default:
					syslog(LOG_ERR, "init_limits: recognized but unknown type of limit timestamp: '%c'", *buf);
					return -1;
				}
				if (ts_ptr == NULL) {
					syslog(LOG_ERR, "%s: wrong order of timestamps", lp->filename);
					return -1;
				}
				ts_ptr->ts_is_set = 1;
				if ( (ts_ptr->ts_offset = ftell(lp->fp)) < 0) {
					syslog(LOG_ERR, "ftell(%s): %m", lp->filename);
					return -1;
				}
				memcpy(&ts_ptr->ts_tm, &curr_tm, sizeof curr_tm); /* init TZ parameters, etc. */
				if (sscanf(buf + 2, "%d.%d.%d/%d:%d:%d", &ts_ptr->ts_tm.tm_year,
				    &ts_ptr->ts_tm.tm_mon, &ts_ptr->ts_tm.tm_mday, &ts_ptr->ts_tm.tm_hour,
				    &ts_ptr->ts_tm.tm_min, &ts_ptr->ts_tm.tm_sec) != 6) {
					syslog(LOG_ERR, "%s: sscanf(\"%s\", \"%%d.%%d.%%d/%%d:%%d:%%d\")", lp->filename, buf + 2);
					return -1;
				}
				ts_ptr->ts_tm.tm_year -= 1900;
				--ts_ptr->ts_tm.tm_mon;
				if (checktm(&ts_ptr->ts_tm) < 0) {
					syslog(LOG_ERR, "%s: wrong values of date in %d line", lp->filename, k + 2);
					return -1;
				}
			}

			if (!ts_start.ts_is_set) {
				syslog(LOG_ERR, "%s: first timestamp must be marked by 's'", lp->filename);
				return -1;
			}

			if (tmcmp(&ts_start.ts_tm, &curr_tm) > 0) {
				syslog(LOG_WARNING, "rule %s limit %s: start_time = %d.%02d.%02d/%02d:%02d:%02d is greater than current time",
				    acp->rulename, lp->limitname,
				    ts_start.ts_tm.tm_year + 1900, ts_start.ts_tm.tm_mon + 1, ts_start.ts_tm.tm_mday,
				    ts_start.ts_tm.tm_hour, ts_start.ts_tm.tm_min, ts_start.ts_tm.tm_sec);
				if (restart_limit(acp, lp) < 0)
					return -1;
			}

			lp->start_time = mktime(&ts_start.ts_tm);

			if (lp->bcnt < old_size) {
				/* limit isn't reached with old "size" */
				if (ts_reach.ts_is_set || ts_expire.ts_is_set ||
				    ts_exec1.ts_is_set || ts_exec2.ts_is_set) {
					syslog(LOG_ERR, "%s: limit isn't reached with old size, wrong timestamps were found",
					    lp->filename);
					return -1;
				}
				if (mode == 1 && ts_zero.ts_is_set) {
					if (tmcmp(&curr_tm, &ts_zero.ts_tm) > 0) {
						syslog(LOG_INFO, "init_limits: rule %s limit %s: old zero_time = %d.%02d.%02d/%02d:%02d:%02d is less than current time",
						    acp->rulename, lp->limitname,
						    ts_zero.ts_tm.tm_year + 1900, ts_zero.ts_tm.tm_mon + 1, ts_zero.ts_tm.tm_mday,
						    ts_zero.ts_tm.tm_hour, ts_zero.ts_tm.tm_min, ts_zero.ts_tm.tm_sec);
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
				if (Ftruncate_Seek(lp->fp, ts_start.ts_offset, lp->filename) < 0)
					return -1;
				if (append_limit_z_ts(lp) < 0)
					return -1;
				if (mode == 1 && lp->zero_time_param.upto != UPTO_NOTSET) {
					if (curr_time > lp->zero_time) {
						struct tm	tmp_tm;

						localtime_r(&lp->zero_time, &tmp_tm);
						if (debug_limit > 0)
							syslog(LOG_INFO, "init_limits: rule %s limit %s: recalculated zero_time = %d.%02d.%02d/%02d:%02d:%02d is less than current time",
							    acp->rulename, lp->limitname,
							    tmp_tm.tm_year + 1900, tmp_tm.tm_mon + 1, tmp_tm.tm_mday,
							    tmp_tm.tm_hour, tmp_tm.tm_min, tmp_tm.tm_sec);
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
			} else {
				/* limit is reached with old "size" */
				if (!ts_reach.ts_is_set) {
					syslog(LOG_ERR, "%s: limit is reached with old size, can't find reach_time line",
					    lp->filename);
					continue;
				}
				lp->reach_time = mktime(&ts_reach.ts_tm);
				if (mode == 1 && ts_expire.ts_is_set) {
					if (tmcmp(&curr_tm, &ts_expire.ts_tm) > 0) {
						if (debug_limit > 0)
							syslog(LOG_INFO, "init_limits: rule %s limit %s: old expire_time = %d.%02d.%02d/%02d:%02d:%02d less than current time",
							    acp->rulename, lp->limitname,
							    ts_expire.ts_tm.tm_year + 1900, ts_expire.ts_tm.tm_mon + 1, ts_expire.ts_tm.tm_mday,
							    ts_expire.ts_tm.tm_hour, ts_expire.ts_tm.tm_min, ts_expire.ts_tm.tm_sec);
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
				if (ts_exec1.ts_is_set) {
					if (Ftruncate_Seek(lp->fp, ts_exec1.ts_offset, lp->filename) < 0)
						return -1;
				} else {
					if (Ftruncate_Seek(lp->fp, ts_reach.ts_offset, lp->filename) < 0)
						return -1;
				}
				if (append_limit_e_ts(lp) < 0)
					return -1;
				if (mode == 1 && lp->expire.time.upto != UPTO_NOTSET) {
					if (curr_time > lp->expire_time) {
						struct tm	tmp_tm;

						localtime_r(&lp->expire_time, &tmp_tm);
						if (debug_limit > 0)
							syslog(LOG_INFO, "init_limits: rule %s limit %s: recalculated expire_time = %d.%02d.%02d/%02d:%02d:%02d is less than current time",
							    acp->rulename, lp->limitname,
							    tmp_tm.tm_year + 1900, tmp_tm.tm_mon + 1, tmp_tm.tm_mday,
							    tmp_tm.tm_hour, tmp_tm.tm_min, tmp_tm.tm_sec);
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
				lp->status = LIMIT_IS_REACHED | LIMIT_EXPIRE_CHECK;
			}
next_limit:
		}
	}
	if (buf)
		free(buf);
	return 0;
}

void
loaded_rules_info(void)
{
	if (nipac == 0)
		syslog(LOG_WARNING, "can't find any rule in config file, hope that's OK");
	else
		syslog(LOG_INFO, "loaded %u accounting rule%s", nipac, nipac == 1 ? "" : "s");
}

int
Ftruncate_Seek(FILE *fp, long length, const char *filename)
{
	if (fflush(fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", filename);
		return -1;
	}
	if (ftruncate(fileno(fp), length) < 0) {
		syslog(LOG_ERR, "ftruncate(%s, %ld): %m", filename, length);
		return -1;
	}
	if (fflush(fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", filename);
		return -1;
	}
	if (fseek(fp, length, SEEK_SET) < 0) {
		syslog(LOG_ERR, "fseek(%s, %ld, SEEK_SET): %m", filename, length);
		return -1;
	}
	return 0;
}

void
sig_chld(int signo)
{
	u_int		i, j;
	int		status, foundchildpid = 0;
	pid_t		childpid;
	struct ipac	*acp = ipac;
	struct limit	*lp;

	wakeup_flag = 1;
	while ( (childpid = waitpid(0, &status, WNOHANG)) > 0) {
		foundchildpid = 0;
		for (i = 0; i < nipac; ++acp, ++i) {
			for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j)
				if (lp->pid == childpid) {
					lp->pid = 0;
					lp->status ^= LIMIT_EXEC;
					foundchildpid = 1;
					break;
				}
			if (foundchildpid > 0)
				break;
		}
		if (foundchildpid == 0)
			syslog(LOG_WARNING, "sig_chld: waitpid return PID %d, such PID isn't expected",
			    childpid);
		sig_chld_flag = 1;
	}
	if (foundchildpid == 0 && childpid < 0 && errno != EINTR)
		syslog(LOG_WARNING, "sig_chld: waitpid: %m");
}

static int
append_limit_z_ts(struct limit *lp)
{
	if (lp->zero_time_param.upto != UPTO_NOTSET) {
		struct tm	tmp_tm;

		lp->zero_time = mk_time_exp(lp->start_time, &lp->zero_time_param);
		localtime_r(&lp->zero_time, &tmp_tm);
		if (append_limit_ts(lp, LIMIT_ZEROED, &tmp_tm) < 0)
			return -1;
	}
	return 0;
}

static int
append_limit_e_ts(struct limit *lp)
{
	if (lp->expire.time.upto != UPTO_NOTSET) {
		struct tm	tmp_tm;

		lp->expire_time = mk_time_exp(lp->reach_time, &lp->expire.time);
		localtime_r(&lp->expire_time, &tmp_tm);
		if (append_limit_ts(lp, LIMIT_EXPIRED, &tmp_tm) < 0)
			return -1;
	}
	return 0;
}
